
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs cpcd_int_bnd() and cpcd_ext_bnd().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "cflds.h"
#include "global.h"

static const sl3_alg_dble cd0={{0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
                               {0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0}};


static void random_flds(int n,sl3_alg_dble **wcd)
{
   int i;

   for (i=0;i<n;i++)
      gauss_dble((double*)(wcd[i]),16*NSPIN);
}


static void set_fld(sl3_alg_dble *cd)
{
   int x0,x1,x2,x3,ix;
   int bo[4];
   sl3_alg_dble *X;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];
               X=cd+ix;

               (*X)=cd0;
               (*X).c1.re=(double)(bo[0]+x0);
               (*X).c2.re=(double)(bo[1]+x1);
               (*X).c3.re=(double)(bo[2]+x2);
               (*X).c4.re=(double)(bo[3]+x3);
            }
         }
      }
   }
}


static void set_bnd(sl3_alg_dble *cd)
{
   int x0,x1,x2,x3;
   int ix,mu,r;
   int N[4],bo[4],x[4];
   sl3_alg_dble *X;

   N[0]=NPROC0*L0;
   N[1]=NPROC1*L1;
   N[2]=NPROC2*L2;
   N[3]=NPROC3*L3;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               if (ix>=(VOLUME/2))
               {
                  x[0]=x0;
                  x[1]=x1;
                  x[2]=x2;
                  x[3]=x3;

                  for (mu=0;mu<4;mu++)
                  {
                     if (iup[ix][mu]>=VOLUME)
                     {
                        r=x[mu];
                        x[mu]+=1;

                        if ((bo[mu]+x[mu])>=N[mu])
                           x[mu]-=N[mu];

                        X=cd+iup[ix][mu];
                        (*X)=cd0;
                        (*X).c1.im=(double)(bo[0]+x[0]);
                        (*X).c2.im=(double)(bo[1]+x[1]);
                        (*X).c3.im=(double)(bo[2]+x[2]);
                        (*X).c4.im=(double)(bo[3]+x[3]);

                        x[mu]=r;
                     }

                     if (idn[ix][mu]>=VOLUME)
                     {
                        r=x[mu];
                        x[mu]-=1;

                        if ((bo[mu]+x[mu])<0)
                           x[mu]+=N[mu];

                        X=cd+idn[ix][mu];
                        (*X)=cd0;
                        (*X).c1.im=(double)(bo[0]+x[0]);
                        (*X).c2.im=(double)(bo[1]+x[1]);
                        (*X).c3.im=(double)(bo[2]+x[2]);
                        (*X).c4.im=(double)(bo[3]+x[3]);

                        x[mu]=r;
                     }
                  }
               }
            }
         }
      }
   }
}


static void check_ext_bnd(sl3_alg_dble *cd)
{
   int x0,x1,x2,x3;
   int ix,mu,r,ie;
   int N[4],bo[4],x[4];
   sl3_alg_dble *X;

   N[0]=NPROC0*L0;
   N[1]=NPROC1*L1;
   N[2]=NPROC2*L2;
   N[3]=NPROC3*L3;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   ie=0;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               if (ix>=(VOLUME/2))
               {
                  x[0]=x0;
                  x[1]=x1;
                  x[2]=x2;
                  x[3]=x3;

                  for (mu=0;mu<4;mu++)
                  {
                     r=x[mu];
                     x[mu]+=1;

                     if ((bo[mu]+x[mu])>=N[mu])
                        x[mu]-=N[mu];

                     X=cd+iup[ix][mu];
                     ie|=((*X).c1.re!=(double)(bo[0]+x[0]));
                     ie|=((*X).c2.re!=(double)(bo[1]+x[1]));
                     ie|=((*X).c3.re!=(double)(bo[2]+x[2]));
                     ie|=((*X).c4.re!=(double)(bo[3]+x[3]));

                     x[mu]=r;
                     x[mu]-=1;

                     if ((bo[mu]+x[mu])<0)
                        x[mu]+=N[mu];

                     X=cd+idn[ix][mu];
                     ie|=((*X).c1.re!=(double)(bo[0]+x[0]));
                     ie|=((*X).c2.re!=(double)(bo[1]+x[1]));
                     ie|=((*X).c3.re!=(double)(bo[2]+x[2]));
                     ie|=((*X).c4.re!=(double)(bo[3]+x[3]));

                     x[mu]=r;
                  }
               }
            }
         }
      }
   }

   error(ie!=0,1,"check_ext_bnd [check1.c]",
         "Incorrectly copied field variables");
}


static int no_nn(int ix)
{
   int mu,n;

   n=0;

   for (mu=0;mu<4;mu++)
   {
      n+=(iup[ix][mu]>=VOLUME);
      n+=(idn[ix][mu]>=VOLUME);
   }

   return n;
}


static void check_int_bnd(sl3_alg_dble *cd)
{
   int x0,x1,x2,x3,ix,n,ie;
   int bo[4];
   sl3_alg_dble *X;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   ie=0;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               X=cd+ix;
               ie|=((*X).c1.re!=(double)(bo[0]+x0));
               ie|=((*X).c2.re!=(double)(bo[1]+x1));
               ie|=((*X).c3.re!=(double)(bo[2]+x2));
               ie|=((*X).c4.re!=(double)(bo[3]+x3));

               if (ix<(VOLUME/2))
                  n=no_nn(ix);
               else
                  n=0;

               ie|=((*X).c1.im!=(double)(n*(bo[0]+x0)));
               ie|=((*X).c2.im!=(double)(n*(bo[1]+x1)));
               ie|=((*X).c3.im!=(double)(n*(bo[2]+x2)));
               ie|=((*X).c4.im!=(double)(n*(bo[3]+x3)));
            }
         }
      }
   }

   error(ie!=0,1,"check_int_bnd [check1.c]",
         "Incorrectly copy-added field variables");
}


int main(int argc,char *argv[])
{
   int my_rank;
   sl3_alg_dble **wcd;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);

      printf("\n");
      printf("Check of the programs cpcd_int_bnd() and cpcd_ext_bnd()\n");
      printf("-------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   start_ranlux(0,12345);
   geometry();
   alloc_wcd(2);
   wcd=reserve_wcd(2);
   random_flds(2,wcd);

   set_fld(wcd[0]);
   cpcd_int_bnd(0,wcd[0]);
   check_ext_bnd(wcd[0]);

   set_fld(wcd[1]);
   set_bnd(wcd[1]);
   cpcd_ext_bnd(0,wcd[1]);
   check_int_bnd(wcd[1]);

   release_wcd();

   if (my_rank==0)
   {
      printf("No errors discovered\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
