/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of ftensor() for plane-wave gauge potentials.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "pmdb.h"
#include "su3fcts.h"
#include "ftensor.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 4

static const int plns[6][2]={{0,1},{0,2},{0,3},{2,3},{3,1},{1,2}};
static double mom[NMAX+1][4],fact[NMAX+1];
static complex_dble fta[3][NMAX+1],ftu[3][NMAX+1],fts[3][NMAX+1];
static sl3_alg_dble Aw[NMAX+1][4];
static su3_dble **ft;


static void set_fact(void)
{
   int i;
   double r;

   r=1.0;

   for (i=0;i<=NMAX;i++)
   {
      r/=(double)(i+1);

      if ((i%2)==0)
         fact[i]=r;
      else
         fact[i]=0.0;
   }
}


static void sinh_poly(int n,complex_dble *z,complex_dble *w)
{
   int k,l,j;

   w[0].re=fact[n-1]*z[0].re;
   w[0].im=fact[n-1]*z[0].im;

   for (k=(n-1);k>0;k--)
   {
      for (l=(n-k);l>=0;l--)
      {
         w[l].re=fact[k-1]*z[l].re;
         w[l].im=fact[k-1]*z[l].im;

         for (j=0;j<l;j++)
         {
            w[l].re+=(z[j].re*w[l-1-j].re-z[j].im*w[l-1-j].im);
            w[l].im+=(z[j].re*w[l-1-j].im+z[j].im*w[l-1-j].re);
         }
      }
   }
}


static void set_ftu(int n,int pln,int *x)
{
   int mu,nu,bc,i;
   double px0,pxk,ph[2];
   complex_dble z,w;

   mu=plns[pln][0];
   nu=plns[pln][1];
   bc=bc_type();

   for (i=0;i<=n;i++)
   {
      px0=mom[i][0]*(double)(x[0]);
      pxk=mom[i][1]*(double)(x[1])+
         mom[i][2]*(double)(x[2])+
         mom[i][3]*(double)(x[3]);

      if (mu==0)
      {
         if (bc==1)
         {
            z.re=-cos(px0+0.5*mom[i][0])*sin(pxk+0.5*mom[i][nu]);
            z.im= cos(px0+0.5*mom[i][0])*cos(pxk+0.5*mom[i][nu]);
         }
         else
         {
            z.re=-sin(px0+0.5*mom[i][0])*cos(pxk+0.5*mom[i][nu]);
            z.im=-sin(px0+0.5*mom[i][0])*sin(pxk+0.5*mom[i][nu]);
         }
      }
      else
      {
         if (bc==1)
         {
            z.re=-sin(px0)*cos(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
            z.im=-sin(px0)*sin(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
         }
         else
         {
            z.re=-cos(px0)*sin(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
            z.im= cos(px0)*cos(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
         }
      }

      ph[0]=2.0*sin(0.5*mom[i][mu]);
      ph[1]=2.0*sin(0.5*mom[i][nu]);

      w.re=ph[0]*Aw[i][nu].c1.re-ph[1]*Aw[i][mu].c1.re;
      w.im=ph[0]*Aw[i][nu].c1.im-ph[1]*Aw[i][mu].c1.im;

      fta[0][i].re=z.re*w.re-z.im*w.im;
      fta[0][i].im=z.re*w.im+z.im*w.re;

      w.re=ph[0]*Aw[i][nu].c2.re-ph[1]*Aw[i][mu].c2.re;
      w.im=ph[0]*Aw[i][nu].c2.im-ph[1]*Aw[i][mu].c2.im;

      fta[1][i].re=z.re*w.re-z.im*w.im;
      fta[1][i].im=z.re*w.im+z.im*w.re;

      fta[2][i].re=-fta[0][i].re-fta[1][i].re;
      fta[2][i].im=-fta[0][i].im-fta[1][i].im;
   }

   for (i=0;i<3;i++)
      sinh_poly(n+1,fta[i],ftu[i]);
}


static void set_fts(int n,int pln,int *x)
{
   int mu,nu,i,j;
   double r;

   if (x[0]>0)
   {
      mu=plns[pln][0];
      nu=plns[pln][1];
      r=0.25;
      set_ftu(n,pln,x);

      for (i=0;i<3;i++)
      {
         for (j=0;j<=n;j++)
         {
            fts[i][j].re=r*ftu[i][j].re;
            fts[i][j].im=r*ftu[i][j].im;
         }
      }

      x[mu]-=1;
      set_ftu(n,pln,x);

      for (i=0;i<3;i++)
      {
         for (j=0;j<=n;j++)
         {
            fts[i][j].re+=r*ftu[i][j].re;
            fts[i][j].im+=r*ftu[i][j].im;
         }
      }

      x[mu]+=1;
      x[nu]-=1;
      set_ftu(n,pln,x);

      for (i=0;i<3;i++)
      {
         for (j=0;j<=n;j++)
         {
            fts[i][j].re+=r*ftu[i][j].re;
            fts[i][j].im+=r*ftu[i][j].im;
         }
      }

      x[mu]-=1;
      set_ftu(n,pln,x);

      for (i=0;i<3;i++)
      {
         for (j=0;j<=n;j++)
         {
            fts[i][j].re+=r*ftu[i][j].re;
            fts[i][j].im+=r*ftu[i][j].im;
         }
      }

      x[mu]+=1;
      x[nu]+=1;
   }
   else
   {
      for (i=0;i<3;i++)
      {
         for (j=0;j<=n;j++)
         {
            fts[i][j].re=0.0;
            fts[i][j].im=0.0;
         }
      }
   }
}


static void sl3_mulc(complex_dble *z,sl3_alg_dble *X,sl3_alg_dble *Y)
{
   (*Y).c1.re=(*z).re*(*X).c1.re-(*z).im*(*X).c1.im;
   (*Y).c1.im=(*z).re*(*X).c1.im+(*z).im*(*X).c1.re;

   (*Y).c2.re=(*z).re*(*X).c2.re-(*z).im*(*X).c2.im;
   (*Y).c2.im=(*z).re*(*X).c2.im+(*z).im*(*X).c2.re;

   (*Y).c3.re=(*z).re*(*X).c3.re-(*z).im*(*X).c3.im;
   (*Y).c3.im=(*z).re*(*X).c3.im+(*z).im*(*X).c3.re;

   (*Y).c4.re=(*z).re*(*X).c4.re-(*z).im*(*X).c4.im;
   (*Y).c4.im=(*z).re*(*X).c4.im+(*z).im*(*X).c4.re;

   (*Y).c5.re=(*z).re*(*X).c5.re-(*z).im*(*X).c5.im;
   (*Y).c5.im=(*z).re*(*X).c5.im+(*z).im*(*X).c5.re;

   (*Y).c6.re=(*z).re*(*X).c6.re-(*z).im*(*X).c6.im;
   (*Y).c6.im=(*z).re*(*X).c6.im+(*z).im*(*X).c6.re;

   (*Y).c7.re=(*z).re*(*X).c7.re-(*z).im*(*X).c7.im;
   (*Y).c7.im=(*z).re*(*X).c7.im+(*z).im*(*X).c7.re;

   (*Y).c8.re=(*z).re*(*X).c8.re-(*z).im*(*X).c8.im;
   (*Y).c8.im=(*z).re*(*X).c8.im+(*z).im*(*X).c8.re;
}


static void set_Aw(void)
{
   int n,mu;

   for (n=0;n<=NMAX;n++)
   {
      for (mu=0;mu<4;mu++)
      {
         sl3_zero(Aw[n]+mu);
         gauss_dble((double*)(Aw[n]+mu),4);
      }
   }

   MPI_Bcast(Aw[0],(NMAX+1)*4*16,MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void set_mom(void)
{
   int n,bc;
   double pi,r[4],p[4];

   bc=bc_type();
   pi=4.0*atan(1.0);

   r[0]=pi/(double)(N0);
   r[1]=2.0*pi/(double)(N1);
   r[2]=2.0*pi/(double)(N2);
   r[3]=2.0*pi/(double)(N3);

   for (n=0;n<=NMAX;n++)
   {
      ranlxd(p,4);

      if (bc==1)
         mom[n][0]=r[0]*floor(p[0]*(double)(N0));
      else
         mom[n][0]=r[0]*(floor(p[0]*(double)(N0))+0.5);

      mom[n][1]=r[1]*floor(p[1]*(double)(N1));
      mom[n][2]=r[2]*floor(p[2]*(double)(N2));
      mom[n][3]=r[3]*floor(p[3]*(double)(N3));
   }

   MPI_Bcast(mom[0],(NMAX+1)*4,MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void random_flds(int n,sl3_alg_dble *ad)
{
   int ix,ifc,i,bc;
   int x0,x1,x2,x3,bo[4],x[4];
   double px0,pxk;
   complex_dble z;
   sl3_alg_dble *A;

   set_Aw();
   set_mom();

   bc=bc_type();
   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               if (ix>=(VOLUME/2))
               {
                  A=ad+(n+1)*8*(ix-(VOLUME/2));

                  x[0]=x0+bo[0];
                  x[1]=x1+bo[1];
                  x[2]=x2+bo[2];
                  x[3]=x3+bo[3];

                  for (ifc=0;ifc<8;ifc++)
                  {
                     if ((ifc==1)&&(x[0]==0))
                        x[0]=N0-1;
                     else if (ifc&0x1)
                        x[ifc/2]-=1;

                     for (i=0;i<=n;i++)
                     {
                        px0=mom[i][0]*(double)(x[0]);
                        pxk=mom[i][1]*(double)(x[1])+
                           mom[i][2]*(double)(x[2])+
                           mom[i][3]*(double)(x[3]);

                        if (ifc<2)
                        {
                           if (bc==1)
                           {
                              z.re=cos(px0+0.5*mom[i][0])*cos(pxk);
                              z.im=cos(px0+0.5*mom[i][0])*sin(pxk);
                           }
                           else
                           {
                              z.re=-sin(px0+0.5*mom[i][0])*sin(pxk);
                              z.im= sin(px0+0.5*mom[i][0])*cos(pxk);
                           }
                        }
                        else
                        {
                           if (bc==1)
                           {
                              z.re=-sin(px0)*sin(pxk+0.5*mom[i][ifc/2]);
                              z.im= sin(px0)*cos(pxk+0.5*mom[i][ifc/2]);
                           }
                           else
                           {
                              z.re=cos(px0)*cos(pxk+0.5*mom[i][ifc/2]);
                              z.im=cos(px0)*sin(pxk+0.5*mom[i][ifc/2]);
                           }
                        }

                        sl3_mulc(&z,Aw[i]+(ifc/2),A);
                        A+=1;
                     }

                     if ((ifc==1)&&(x[0]==(N0-1)))
                        x[0]=0;
                     else if (ifc&0x1)
                        x[ifc/2]+=1;
                  }
               }
            }
         }
      }
   }
}


static void set_ft(int n,sl3_alg_dble *ad)
{
   int ix,ifc;
   su3_dble **wud,*ud;

   wud=reserve_wud(n+1);
   ud=wud[0];

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         pexp_sl3_alg(n+1,ad,ud);
         ad+=(n+1);
         ud+=(n+1);
      }
   }

   ft=ftensor(n,wud[0]);
   release_wud();
}


static double chk_ft1(int n)
{
   int ix,pln,i;
   double d,dmax;
   su3_dble *ftn;

   dmax=0.0;

   for (pln=0;pln<6;pln++)
   {
      ftn=ft[pln];

      for (ix=0;ix<VOLUME;ix++)
      {
         for (i=0;i<=n;i++)
         {
            d=fabs((*ftn).c12.re)+fabs((*ftn).c12.im);
            if (d>dmax)
               dmax=d;

            d=fabs((*ftn).c21.re)+fabs((*ftn).c21.im);
            if (d>dmax)
               dmax=d;

            d=fabs((*ftn).c13.re)+fabs((*ftn).c13.im);
            if (d>dmax)
               dmax=d;

            d=fabs((*ftn).c31.re)+fabs((*ftn).c31.im);
            if (d>dmax)
               dmax=d;

            d=fabs((*ftn).c23.re)+fabs((*ftn).c23.im);
            if (d>dmax)
               dmax=d;

            d=fabs((*ftn).c32.re)+fabs((*ftn).c32.im);
            if (d>dmax)
               dmax=d;

            ftn+=1;
         }
      }
   }

   d=dmax;
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(&dmax,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   return dmax;
}


static double chk_ft2(int n)
{
   int ix,pln,i;
   int x0,x1,x2,x3,bo[4],x[4];
   double d,dmax;
   su3_dble *ftn;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   dmax=0.0;

   for (pln=0;pln<6;pln++)
   {
      for (x0=0;x0<L0;x0++)
      {
         for (x1=0;x1<L1;x1++)
         {
            for (x2=0;x2<L2;x2++)
            {
               for (x3=0;x3<L3;x3++)
               {
                  ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

                  x[0]=x0+bo[0];
                  x[1]=x1+bo[1];
                  x[2]=x2+bo[2];
                  x[3]=x3+bo[3];

                  set_fts(n,pln,x);
                  ftn=ft[pln]+ix*(n+1);

                  for (i=0;i<=n;i++)
                  {
                     d= fabs(ftn[i].c11.re-fts[0][i].re)+
                        fabs(ftn[i].c11.im-fts[0][i].im);
                     if (d>dmax)
                        dmax=d;

                     d= fabs(ftn[i].c22.re-fts[1][i].re)+
                        fabs(ftn[i].c22.im-fts[1][i].im);
                     if (d>dmax)
                        dmax=d;

                     d= fabs(ftn[i].c33.re-fts[2][i].re)+
                        fabs(ftn[i].c33.im-fts[2][i].im);
                     if (d>dmax)
                        dmax=d;
                  }
               }
            }
         }
      }
   }

   d=dmax;
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(&dmax,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   return dmax;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n;
   double dev[2];
   sl3_alg_dble **wad;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);

      printf("\n");
      printf("Check of ftensor() for plane-wave gauge potentials\n");
      printf("--------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");
      error_root((bc==0)||(sscanf(argv[bc+1],"%d",&bc)!=1),1,
                 "main [check2.c]","Syntax: check2 -bc <type>");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,1234);
   geometry();

   set_fact();
   alloc_wad(NMAX+1);
   alloc_wud(NMAX+1);

   for (n=0;n<=NMAX;n++)
   {
      wad=reserve_wad(n+1);
      random_flds(n,wad[0]);
      set_ft(n,wad[0]);
      dev[0]=chk_ft1(n);
      dev[1]=chk_ft2(n);

      if (my_rank==0)
      {
         printf("n = %d, deviation (off-diagonal) = %.1e, "
                "deviation (diagonal) = %.1e\n",n,dev[0],dev[1]);
         fflush(flog);
      }

      release_wad();
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
