
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Visual inspection of the structure returned by vertices().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "utils.h"
#include "diagrams.h"


static theory_t theory(char *name)
{
   if (strcmp("PHI4",name)==0)
      return PHI4;
   else if (strcmp("YM",name)==0)
      return YM;
   else if (strcmp("QCD",name)==0)
      return QCD;
   else if (strcmp("LYM",name)==0)
      return LYM;
   else if (strcmp("LQCD",name)==0)
      return LQCD;
   else
      error(1,1,"theory [check1.c]","Unknown theory");

   return THEORIES;
}


static void print_vertex(int order,vertex_t *v)
{
   int i,j;

   error(order!=(*v).order,1,"print_vertex [check1.c]",
         "Unexpected order of the vertex");

   if ((*v).type==SOURCE)
      printf("SOURCE vertex: ");
   else if ((*v).type==EXTERNAL)
      printf("EXTERNAL vertex: ");
   else if ((*v).type==VS)
      printf("VS vertex: ");
   else if ((*v).type==VG)
      printf("VG vertex: ");
   else if ((*v).type==VFP)
      printf("VFP vertex: ");
   else if ((*v).type==VQG)
      printf("VQG vertex: ");
   else if ((*v).type==VGM)
      printf("VGM vertex: ");
   else if ((*v).type==VSM)
      printf("VSM vertex: ");
   else if ((*v).type==VQM)
      printf("VQM vertex: ");
   else if ((*v).type==VGF)
      printf("VGF vertex: ");
   else
      error(1,1,"print_vertex [check1.c]","Unknown vertex type");

   j=0;

   for (i=0;i<(int)(LTYPES);i++)
   {
      if (((int)(SCALAR)==i)&&((*v).nl[i]>0))
      {
         if (j>0)
            printf(",");
         printf("%d SCALAR",(*v).nl[i]);
         j=1;
      }
      else if (((int)(GLUON)==i)&&((*v).nl[i]>0))
      {
         if (j>0)
            printf(",");
         printf("%d GLUON",(*v).nl[i]);
         j=1;
      }
      else if (((int)(GHOST)==i)&&((*v).nl[i]>0))
      {
         if (j>0)
            printf(",");
         printf("%d GHOST",(*v).nl[i]);
         j=1;
      }
      else if (((int)(ANTIGHOST)==i)&&((*v).nl[i]>0))
      {
         if (j>0)
            printf(",");
         printf("%d ANTIGHOST",(*v).nl[i]);
         j=1;
      }
      else if (((int)(QUARK)==i)&&((*v).nl[i]>0))
      {
         if (j>0)
            printf(",");
         printf("%d QUARK",(*v).nl[i]);
         j=1;
      }
      else if (((int)(ANTIQUARK)==i)&&((*v).nl[i]>0))
      {
         if (j>0)
            printf(",");
         printf("%d ANTIQUARK",(*v).nl[i]);
         j=1;
      }
   }

   printf("\n");
}


int main(void)
{
   int order,i,j;
   char name[5];
   theory_t th;
   vlist_t vl;

   printf("\n");
   printf("List of vertices\n");
   printf("----------------\n\n");

   while(1)
   {
      printf("Specify theory and maximal order: ");
      (void)scanf("%4s %d",name,&order);

      th=theory(name);
      vl=vertices(th,order);

      printf("\n");
      printf("Content of theory structure:\n\n");
      printf("Maximal order = %d\n",vl.order);

      for (i=0;i<=vl.order;i++)
      {
         printf("Order %d: No of vertices = %d\n",i,vl.nv[i]);

         for (j=0;j<vl.nv[i];j++)
            print_vertex(i,vl.v[i]+j);
      }

      printf("\n");
   }

   exit(0);
}
