
/*******************************************************************************
*
* File table1.c
*
* Copyright (C) 2014,2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Table of the normalization factor of the gradient-flow coupling (see
* modules/nompi/gbar/kfact.c and doc/gfnorm.pdf).
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "gbar.h"

static int Ts,Ls,bcs,Ns;
static double c1s,ts;
static char type[NAME_SIZE];


static void read_parms(void)
{
   FILE *fin=NULL;

   fin=fopen("table1.in","r");
   error(fin==NULL,1,"read_parms [table1.c]",
         "Unable to open input parameter file");

   find_section(fin,"Lattice");
   read_line(fin,"T","%d",&Ts);
   read_line(fin,"L","%d",&Ls);
   read_line(fin,"bc","%d",&bcs);

   find_section(fin,"Theory");
   read_line(fin,"N","%d",&Ns);
   read_line(fin,"c1","%lf\n",&c1s);

   find_section(fin,"Coupling");
   read_line(fin,"type","%s\n",type);
   read_line(fin,"t","%lf\n",&ts);

   error((Ts<4)||(Ls<4)||(bcs<1)||(bcs>2)||(Ns<=1)||(c1s>=0.125)||(ts<=0.0),1,
         "read_parms [table1.c]","Improper choice of input parameters");

   fclose(fin);
}


static void print_table(FILE *fdat)
{
   int x0,n;
   double c0,k;

   fprintf(fdat,"#\n");
   fprintf(fdat,"# Normalization factor k_%s of the gradient-flow coupling\n",
           type);
   fprintf(fdat,"#\n");
   fprintf(fdat,"# N = %d, T = %d, L = %d\n",Ns,Ts,Ls);

   if (bcs==1)
      fprintf(fdat,"# SF boundary conditions\n");
   else
      fprintf(fdat,"# Open-SF boundary conditions\n");

   c0=1.0-8.0*c1s;
   n=fdigits(c0);
   fprintf(fdat,"# c0 = %.*f, ",IMAX(n,1),c0);
   n=fdigits(c1s);
   fprintf(fdat,"c1 = %.*f, ",IMAX(n,1),c1s);
   n=fdigits(ts);
   fprintf(fdat,"t = %.*f\n",IMAX(n,1),ts);
   fprintf(fdat,"#\n");

   fprintf(fdat,"#   x0       k            1/k\n");

   for (x0=1;x0<Ts;x0++)
   {
      k=kfact(type,Ns,Ts,Ls,bcs,c1s,x0,ts);

      fprintf(fdat,"  %3d   %12.6e   %12.6e\n",x0,k,1.0/k);
   }
}


int main(int argc,char *argv[])
{
   FILE *fdat;

   read_parms();
   fdat=fopen("plots/table1.dat","w");
   error(fdat==NULL,1,"main [table1.c]","Unable to open output file");
   print_table(fdat);
   fclose(fdat);

   printf("\n");
   printf("Normalization factor k_%s of the gradient-flow coupling\n\n",
          type);
   printf("Table printed to file plots/table1.dat\n\n");
   exit(0);
}
