
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of line lists.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"

static int m,*nv,*nl,**v,**ne,(*p)[2],(*ps)[2];
static int **iv;


static int idx(int i,int lab)
{
   int j;

   for (j=0;j<nv[i];j++)
   {
      if (lab==v[i][j])
         return j;
   }

   return nv[i];
}


static void save_p(void)
{
   int i,j,k;

   k=0;

   for (i=0;i<m;i++)
   {
      for (j=0;j<nl[i];j++)
      {
         ps[k][0]=p[k][0];
         ps[k][1]=p[k][1];
         k+=1;
      }
   }
}


static void check_order(void)
{
   int i,j,k;
   int i0,i1,j0,j1;

   for (i=(m-1);i>=0;i--)
   {
      k=0;

      for (j=0;j<i;j++)
         k+=nl[j];

      for (j=0;j<nl[i];j++)
      {
         i0=idx(i,ps[k][0]);
         i1=idx(i,ps[k][1]);
         j0=idx(i,p[k][0]);
         j1=idx(i,p[k][1]);

         if ((i0<j0)||((i0==j0)&&(i1<j1)))
            return;
         else if ((i0>j0)||((i0==j0)&&(i1>j1)))
            error(1,1,"check_order [check5.c]",
                  "line lists are not properly ordered");
         k+=1;
      }
   }

   error(1,2,"check_order [check5.c]",
         "line lists are not properly ordered");
}


static void check_lns(void)
{
   int i,j,i1,i2,ie;
   int (*r)[2];

   ie=0;
   r=p;

   for (i=0;i<m;i++)
   {
      for (j=0;j<nv[i];j++)
         iv[i][j]=0;

      for (j=0;j<nl[i];j++)
      {
         i1=idx(i,r[0][0]);
         i2=idx(i,r[0][1]);

         if ((i1<nv[i])&&(i2<nv[i]))
         {
            iv[i][i1]+=1;
            iv[i][i2]+=1;

            ie|=(i1>i2);
         }
         else
            ie=1;

         if (j>0)
         {
            ie|=(i1<idx(i,r[-1][0]));
            ie|=((i1==idx(i,r[-1][0]))&&(i2<idx(i,r[-1][1])));
         }

         r+=1;
      }

      for (j=0;j<nv[i];j++)
         ie|=(iv[i][j]>ne[i][j]);
   }

   error(ie!=0,1,"check_lns [check5.c]","Incorrect line list");
}


static void check_nctr(long na)
{
   int i,j,n;
   long f0,f1;

   f0=1L;

   for (i=0;i<m;i++)
   {
      n=0;

      for (j=0;j<nv[i];j++)
         n+=ne[i][j];

      f1=binomial(n,2*nl[i]);

      if (f0<=(LONG_MAX/f1))
         f0*=f1;
      else
         error(1,1,"check_nctr [check5.c]",
               "Total number of contractions exceeds LONG_MAX");

      f1=factorial2(2*nl[i]-1);

      if (f0<=(LONG_MAX/f1))
         f0*=f1;
      else
         error(1,1,"check_nctr [check5.c]",
               "Total number of contractions exceeds LONG_MAX");
   }

   error(na!=f0,1,"check_nctr [check5.c]",
         "Incorrect total number of contractions");
}


int main(void)
{
   int nvt,nlt,i,j,k;
   long na,nc,np,npm;

   printf("\n");
   printf("Generation of line lists\n");
   printf("------------------------\n\n");

   while(1)
   {
      printf("Specify number of subsequences of vertices: ");
      (void)scanf("%d",&m);

      nv=malloc(2*m*sizeof(*nv));
      v=malloc(3*m*sizeof(*v));
      error((nv==NULL)||(v==NULL),1,"main [check5.c]",
            "Unable to allocate auxiliary arrays");

      nl=nv+m;
      ne=v+m;
      iv=ne+m;

      printf("Specify lengths of subsequences: ");

      for (i=0;i<m;i++)
         (void)scanf("%d",nv+i);

      printf("Specify number of contractions in each subsequence: ");

      for (i=0;i<m;i++)
         (void)scanf("%d",nl+i);

      printf("\n");
      nvt=0;
      nlt=0;

      for (i=0;i<m;i++)
      {
         nvt+=nv[i];
         nlt+=nl[i];
      }

      v[0]=malloc(3*nvt*sizeof(**v));
      p=malloc(2*nlt*sizeof(*p));
      error((v[0]==NULL)||(p==NULL),2,"main [check5.c]",
            "Unable to allocate auxiliary arrays");
      ne[0]=v[0]+nvt;
      iv[0]=ne[0]+nvt;
      ps=p+nlt;

      for (i=0;i<m;i++)
      {
         if (i>0)
         {
            v[i]=v[i-1]+nv[i-1];
            ne[i]=ne[i-1]+nv[i-1];
            iv[i]=iv[i-1]+nv[i-1];
         }

         printf("Specify vertex labels (subsequence %d, %d elements): ",
                i,nv[i]);

         for (j=0;j<nv[i];j++)
            (void)scanf("%d",v[i]+j);

         printf("Specify no of external lines (subsequence %d, %d elements): ",
                i,nv[i]);

         for (j=0;j<nv[i];j++)
            (void)scanf("%d",ne[i]+j);

         printf("\n");
      }

      printf("Line lists:\n");

      na=0L;
      np=0L;
      npm=init_lnlist(m,nv,nl,v,ne);

      printf("Total number of line lists = %ld\n",npm);

      while (np<npm)
      {
         np=next_lnlist(p,&nc);
         printf("np = %3ld, ",np);
         k=0;

         for (i=0;i<m;i++)
         {
            if (i==0)
               printf("[");
            else
               printf(",[");

            for (j=0;j<nl[i];j++)
            {
               if (j==0)
                  printf("(%d,%d)",p[k][0],p[k][1]);
               else
                  printf(",(%d,%d)",p[k][0],p[k][1]);
               k+=1;
            }

            printf("]");
         }

         printf(", nc = %ld\n",nc);

         if (na<=(LONG_MAX-nc))
            na+=nc;
         else
            error(1,1,"main [check5.c]",
                  "Total number of contractions exceeds LONG_MAX");

         check_lns();
         if (np>1L)
            check_order();
         save_p();
      }

      printf("Total number of contractions = %ld\n",na);
      printf("Reduction factor = %ld\n",(na+npm/2)/npm);
      check_nctr(na);

      printf("\n");
      printf("Wrapping around:\n");
      np=0L;

      while (np<(npm/2))
      {
         np=next_lnlist(p,&nc);
         printf("np = %3ld, ",np);
         k=0;

         for (i=0;i<m;i++)
         {
            if (i==0)
               printf("[");
            else
               printf(",[");

            for (j=0;j<nl[i];j++)
            {
               if (j==0)
                  printf("(%d,%d)",p[k][0],p[k][1]);
               else
                  printf(",(%d,%d)",p[k][0],p[k][1]);
               k+=1;
            }

            printf("]");
         }

         printf(", nc = %ld\n",nc);

         check_lns();
      }

      printf("...\n\n");
      stop_lnlist();
   }

   exit(0);
}
