
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of sym_prod().
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "utils.h"
#include "vxfcts.h"


int main(void)
{
   int n,nt,count,np;
   double t1,t2,dt;
   su3_dble *X,*PX;

   printf("\n");
   printf("Timing of sym_prod()\n");
   printf("--------------------\n\n");

#if (defined AVX)
   printf("Using AVX instructions\n");
#elif (defined x64)
   printf("Using SSE3 instructions and up to 16 xmm registers\n");
#endif

   printf("Measurement made with all data in cache\n\n");
   rlxd_init(1,23456);
   X=NULL;

   while (1)
   {
      printf("Specify number of factors: ");
      (void)(scanf("%d",&n));

      if (X!=NULL)
         afree(X);

      X=amalloc((n+(1<<n))*sizeof(*X),4);
      error(X==NULL,1,"main [time1.c]",
            "Unable to allocate matrix array");
      PX=X+n;
      gauss_dble((double*)(X),18*n);

      sym_prod(0,n,X,PX);
      np=n*((1<<(n-1))-1);
      nt=(int)(4.0e6)/np+1;
      dt=0.0;

      while (dt<2.0)
      {
         t1=(double)clock();
         for (count=0;count<nt;count++)
            sym_prod(0,n,X,PX);
         t2=(double)clock();
         dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
         nt*=2;
      }

      dt*=2.0e6/((double)(nt)*(double)(np));

      printf("The time per matrix multiplication is: %.2e usec\n\n",dt);
   }

   exit(0);
}
