
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program AFPn().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "sl3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "linalg.h"
#include "aflds.h"
#include "cflds.h"
#include "vxfcts.h"
#include "global.h"

#define NMAX 4

extern complex_dble AFPn_action(int n,sl3_alg_dble *ad,
                                sl3_alg_dble *cbd,sl3_alg_dble *cd);

static sl3_alg_dble *vad[NMAX-1];
static su3_dble vd ALIGNED16;
static su3_dble X ALIGNED16;
static su3_dble Y ALIGNED16;


static void new_afld(int n,sl3_alg_dble **ad)
{
   int i;

   for (i=0;i<n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)));

      set_adbc(ad[i]);
   }
}


static void new_cfld(int n,sl3_alg_dble **cd)
{
   int i;

   for (i=0;i<n;i++)
   {
      gauss_dble((double*)(cd[i]),16*NSPIN);
      set_cdbc(cd[i]);
   }
}


static void global_transform_ad(int n,sl3_alg_dble **ad)
{
   int i,ix,ifc;
   sl3_alg_dble *fld;

   for (i=0;i<n;i++)
   {
      fld=ad[i];

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         for (ifc=0;ifc<8;ifc++)
         {
            sl3_to_su3(fld,&X);
            su3xsu3(&vd,&X,&Y);
            su3xsu3dag(&Y,&vd,&X);
            su3_to_sl3(&X,fld);
            fld+=1;
         }
      }
   }
}


static void global_transform_cd(int n,sl3_alg_dble **cd)
{
   int i,ix;
   sl3_alg_dble *fld;

   for (i=0;i<n;i++)
   {
      fld=cd[i];

      for (ix=0;ix<VOLUME;ix++)
      {
         sl3_to_su3(fld,&X);
         su3xsu3(&vd,&X,&Y);
         su3xsu3dag(&Y,&vd,&X);
         su3_to_sl3(&X,fld);
         fld+=1;
      }
   }
}


static void random_vec(int *svec)
{
   int mu,bs[4];
   double r[4];

   bs[0]=NPROC0*L0;
   bs[1]=NPROC1*L1;
   bs[2]=NPROC2*L2;
   bs[3]=NPROC3*L3;
   svec[0]=0;
   svec[1]=0;
   svec[2]=0;
   svec[3]=0;

   while ((abs(svec[1])+abs(svec[2])+abs(svec[3]))==0)
   {
      ranlxd(r,4);

      for (mu=1;mu<4;mu++)
      {
         svec[mu]=(int)((double)(bs[mu])*r[mu]);
         if (svec[mu]>(bs[mu]/2))
            svec[mu]-=bs[mu];
      }
   }

   MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);
}


static void shift_aflds(int n,int *s,sl3_alg_dble **ad)
{
   int i;

   for (i=0;i<n;i++)
      shift_ad(s,ad[i]);
}


static void shift_cflds(int n,int *s,sl3_alg_dble **cd)
{
   int i;

   for (i=0;i<n;i++)
      shift_cd(s,cd[i]);
}


static double fact(int n)
{
   int i;
   double f;

   f=1.0;

   for (i=1;i<=n;i++)
      f*=(double)(i);

   return f;
}


int main(int argc,char *argv[])
{
   int my_rank,bc;
   int n,i,s[4];
   double d,dmax,r;
   complex_dble z,w;
   sl3_alg_dble **wad,**wcd,*wa1,*wa2;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);

      printf("\n");
      printf("Check of the program AFPn()\n");
      printf("---------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check3.c]",
                    "Syntax: check3 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();
   alloc_wad(NMAX+4);
   alloc_wcd(2);
   wad=reserve_wad(NMAX+2);
   wcd=reserve_wcd(2);

   if (my_rank==0)
      printf("Global gauge invariance:\n");

   for (n=1;n<=NMAX;n+=(1+(n!=1)))
   {
      random_su3_dble(&vd);
      MPI_Bcast(&vd,18,MPI_DOUBLE,0,MPI_COMM_WORLD);
      new_afld(n-1,wad);
      new_cfld(2,wcd);
      AFPn(n,wad,wcd[0],wcd[1],wad[n-1]);
      global_transform_ad(n,wad);
      global_transform_cd(2,wcd);
      AFPn(n,wad,wcd[0],wcd[1],wad[n]);
      mulr_add_calg(4*VOLUME,wad[n],wad[n-1],-1.0);
      d=norm_square_calg(4*VOLUME,1,wad[n])/
         norm_square_calg(4*VOLUME,1,wad[n-1]);

      if (my_rank==0)
      {
         printf("n=%d, deviation = %.1e\n",n,sqrt(d));
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Translation invariance:\n");
   }

   for (n=1;n<=NMAX;n+=(1+(n!=1)))
   {
      new_afld(n-1,wad);
      new_cfld(2,wcd);
      AFPn(n,wad,wcd[0],wcd[1],wad[n-1]);
      random_vec(s);
      shift_aflds(n,s,wad);
      shift_cflds(2,s,wcd);
      AFPn(n,wad,wcd[0],wcd[1],wad[n]);
      mulr_add_calg(4*VOLUME,wad[n],wad[n-1],-1.0);
      d=norm_square_calg(4*VOLUME,1,wad[n])/
         norm_square_calg(4*VOLUME,1,wad[n-1]);

      if (my_rank==0)
      {
         printf("n=%d, s= % 3d,% 3d,% 3d,% 3d, deviation = %.1e\n",
                n,s[0],s[1],s[2],s[3],sqrt(d));
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Permutation symmetry:\n");
   }

   for (n=2;n<=NMAX;n+=2)
   {
      new_afld(n-1,wad);
      new_cfld(2,wcd);
      AFPn(n,wad,wcd[0],wcd[1],wad[n-1]);
      dmax=0.0;

      for (i=0;i<(n-2);i++)
      {
         wa1=wad[i];
         wa2=wad[i+1];
         wad[i]=wa2;
         wad[i+1]=wa1;

         AFPn(n,wad,wcd[0],wcd[1],wad[n]);
         mulr_add_calg(4*VOLUME,wad[n],wad[n-1],-1.0);
         d=norm_square_calg(4*VOLUME,1,wad[n])/
            norm_square_calg(4*VOLUME,1,wad[n-1]);
         if (d>dmax)
            dmax=d;

         wad[i]=wa1;
         wad[i+1]=wa2;
      }

      new_afld(1,wad+n+1);
      wa1=wad[0];
      wad[0]=wad[n+1];
      AFPn(n,wad,wcd[0],wcd[1],wad[n]);
      wad[0]=wa1;
      assign_calgdag2calg(4*VOLUME,wad[0],wad[0]);
      assign_calgdag2calg(4*VOLUME,wad[n+1],wad[n+1]);
      z=scalar_prod_calg(4*VOLUME,1,wad[0],wad[n]);
      w=scalar_prod_calg(4*VOLUME,1,wad[n+1],wad[n-1]);
      w.re-=z.re;
      w.im-=z.im;
      d=(w.re*w.re+w.im*w.im)/(z.re*z.re+z.im*z.im);
      if (d>dmax)
         dmax=d;

      if (my_rank==0)
      {
         printf("n=%d, deviation = %.1e\n",n,sqrt(dmax));
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Multi-linearity:\n");
   }

   for (n=2;n<=NMAX;n+=2)
   {
      new_afld(n,wad);
      new_cfld(2,wcd);
      AFPn(n,wad,wcd[0],wcd[1],wad[n]);
      AFPn(n,wad+1,wcd[0],wcd[1],wad[n+1]);
      z.re=0.7896;
      z.im=-1.543;
      mulc_add_calg(4*VOLUME,wad[n],wad[n+1],z);
      mulc_add_calg(4*VOLUME,wad[0],wad[n-1],z);
      AFPn(n,wad,wcd[0],wcd[1],wad[n+1]);

      mulr_add_calg(4*VOLUME,wad[n],wad[n+1],-1.0);
      d=norm_square_calg(4*VOLUME,1,wad[n])/
         norm_square_calg(4*VOLUME,1,wad[n+1]);

      if (my_rank==0)
      {
         printf("n=%d, deviation = %.1e\n",n,sqrt(d));
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Comparison with action:\n");
   }

   for (n=1;n<=NMAX;n+=(1+(n!=1)))
   {
      new_afld(1,wad);
      new_cfld(2,wcd);
      assign_calgdag2calg(4*VOLUME,wad[0],wad[1]);

      for (i=0;i<(n-1);i++)
         vad[i]=wad[0];

      AFPn(n,vad,wcd[0],wcd[1],wad[2]);
      z=scalar_prod_calg(4*VOLUME,1,wad[1],wad[2]);
      r=1.0/fact(n);
      z.re*=r;
      z.im*=r;
      w=AFPn_action(n,wad[0],wcd[0],wcd[1]);
      w.re-=z.re;
      w.im-=z.im;
      d=(w.re*w.re+w.im*w.im)/(z.re*z.re+z.im*z.im);

      if (my_rank==0)
      {
         printf("n=%d, deviation = %.1e\n",n,sqrt(d));
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
