
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the programs AFPn(), CFPn() and CbFPn().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "vxfcts.h"
#include "global.h"

#define NMAX 4

static int my_rank;


static void new_afld(int n,sl3_alg_dble **ad)
{
   int i;

   for (i=0;i<n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)));

      set_adbc(ad[i]);
   }
}


static void new_cfld(int n,sl3_alg_dble **cd)
{
   int i;

   for (i=0;i<n;i++)
   {
      gauss_dble((double*)(cd[i]),16*NSPIN);
      set_cdbc(cd[i]);
   }
}


static double wt_AFPn(int k)
{
   int nflds,n,nmax,i,ib;
   double wt1,wt2,wdt,wtav;
   sl3_alg_dble **wad,**wcd;

   nflds=(int)((8*1024*1024)/((4*k+2)*16*VOLUME*sizeof(double)))+1;
   alloc_wad(nflds*k+1);
   alloc_wcd(nflds*2);
   wad=reserve_wad(nflds*k);
   wcd=reserve_wcd(nflds*2);

   new_afld(nflds*k,wad);
   new_cfld(nflds*2,wcd);
   AFPn(k,wad,wcd[0],wcd[1],wad[k-1]);
   nmax=1;

   for (ib=0;ib<1;nmax*=2)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i++)
            AFPn(k,wad+i*k,wcd[i*2],wcd[i*2+1],wad[i*k+k-1]);
      }

      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)(nmax*nflds);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   release_wad();
   release_wcd();

   return wtav;
}


static double wt_CFPn(int k)
{
   int nflds,n,nmax,i,ib;
   double wt1,wt2,wdt,wtav;
   sl3_alg_dble **wad,**wcd;

   nflds=(int)((8*1024*1024)/((4*k+2)*16*VOLUME*sizeof(double)))+1;
   alloc_wad(nflds*k+1);
   alloc_wcd(nflds*2);
   wad=reserve_wad(nflds*k);
   wcd=reserve_wcd(nflds*2);

   new_afld(nflds*k,wad);
   new_cfld(nflds*2,wcd);
   CFPn(k,wad,wcd[0],wcd[1]);
   nmax=1;

   for (ib=0;ib<1;nmax*=2)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i++)
            CFPn(k,wad+i*k,wcd[i*2],wcd[i*2+1]);
      }

      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)(nmax*nflds);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   release_wad();
   release_wcd();

   return wtav;
}


static double wt_CbFPn(int k)
{
   int nflds,n,nmax,i,ib;
   double wt1,wt2,wdt,wtav;
   sl3_alg_dble **wad,**wcd;

   nflds=(int)((8*1024*1024)/((4*k+2)*16*VOLUME*sizeof(double)))+1;
   alloc_wad(nflds*k+1);
   alloc_wcd(nflds*2);
   wad=reserve_wad(nflds*k);
   wcd=reserve_wcd(nflds*2);

   new_afld(nflds*k,wad);
   new_cfld(nflds*2,wcd);
   CbFPn(k,wad,wcd[0],wcd[1]);
   nmax=1;

   for (ib=0;ib<1;nmax*=2)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i++)
            CbFPn(k,wad+i*k,wcd[i*2],wcd[i*2+1]);
      }

      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)(nmax*nflds);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   release_wad();
   release_wcd();

   return wtav;
}


int main(int argc,char *argv[])
{
   int bc,n;
   double wdt;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time2.log","w",stdout);

      printf("\n");
      printf("Timing of the programs AFPn(), CFPn() and CbFPn()\n");
      printf("-------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [time2.c]",
                    "Syntax: time2 [-bc <type>]");

      if (NPROC>1)
         printf("There are %d MPI processes\n",NPROC);
      else
         printf("There is 1 MPI process\n");

      if ((64*VOLUME*sizeof(double))<(4096*1024))
         printf("The local size of a complex gauge potential is %d KB\n",
                (int)((64*VOLUME*sizeof(double))/(1024)));
      else
         printf("The local size of a complex gauge potential is %d MB\n",
                (int)((64*VOLUME*sizeof(double))/(1024*1024)));

#if (defined x64)
#if (defined AVX)
      printf("Using AVX instructions\n");
#else
      printf("Using SSE3 instructions and 16 xmm registers\n");
#endif
#if (defined P3)
      printf("Assuming SSE prefetch instructions fetch 32 bytes\n");
#elif (defined PM)
      printf("Assuming SSE prefetch instructions fetch 64 bytes\n");
#elif (defined P4)
      printf("Assuming SSE prefetch instructions fetch 128 bytes\n");
#else
      printf("SSE prefetch instructions are not used\n");
#endif
#endif
      printf("\n");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   for (n=1;n<=NMAX;n+=(1+(n!=1)))
   {
      wdt=wt_AFPn(n);

      if (my_rank==0)
      {
         printf("n=%d\n",n);
         printf("Time per application of AFPn(): %.1e sec\n",wdt);
         printf("Time per lattice point: %4.3f usec\n\n",
                1.0e6*wdt/(double)(VOLUME));
         fflush(flog);
      }

      wdt=wt_CFPn(n);

      if (my_rank==0)
      {
         printf("n=%d\n",n);
         printf("Time per application of CFPn(): %.1e sec\n",wdt);
         printf("Time per lattice point: %4.3f usec\n\n",
                1.0e6*wdt/(double)(VOLUME));
         fflush(flog);
      }

      wdt=wt_CbFPn(n);

      if (my_rank==0)
      {
         printf("n=%d\n",n);
         printf("Time per application of CbFPn(): %.1e sec\n",wdt);
         printf("Time per lattice point: %4.3f usec\n\n",
                1.0e6*wdt/(double)(VOLUME));
         fflush(flog);
      }
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
