
/*******************************************************************************
*
* File pmdb.h
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
*******************************************************************************/

#ifndef PMDB_H
#define PMDB_H

#ifndef DIAGRAMS_H
#include "diagrams.h"
#endif

#ifndef SL3_H
#include "sl3.h"
#endif

typedef enum
{
   EXP,SIN,COS,
   DFT_TYPES
} dft_type_t;

typedef struct
{
   dft_type_t type;
   int n,b,c;
   int *r;
   complex_dble *w;
   complex_dble *wb,*wc,*iwb,*iwc;
} dft_parms_t;

typedef struct
{
   int csize;
   int ny[4],nf[4];
   int *nx[4],*mf[4];
   dft_parms_t *dp[4];
} dft4d_parms_t;

typedef enum
{
   C0,LAMBDA,MASS,CSW,CG,CGP,CF,CFP,
   PM_TYPES
} pm_type_t;

typedef struct
{
   int order,iroot;
   int opv[(int)(PM_TYPES)];
   double *pv[(int)(PM_TYPES)];
} ispt_parms_t;

typedef struct
{
   theory_t th;
   int bc,Nf;
} th_parms_t;

/* DFT_PARMS_C */
extern int set_dft_parms(dft_type_t type,int n,int b,int c);
extern dft_parms_t *dft_parms(int id);

/* DFT4D_PARMS_C */
extern int set_dft4d_parms(int *idp,int *nx,int csize);
extern dft4d_parms_t *dft4d_parms(int id);

/* ISPT_PARMS_C */
extern ispt_parms_t set_ispt_parms(int order,int iroot,int *opv,double **pv);
extern ispt_parms_t ispt_parms(void);
extern void print_ispt_parms(void);
extern void write_ispt_parms(FILE *fdat);
extern void check_ispt_parms(FILE *fdat);

/* TH_PARMS_C */
extern th_parms_t set_th_parms(theory_t th,int bc,int Nf);
extern th_parms_t th_parms(void);
extern void print_th_parms(void);
extern void write_th_parms(FILE *fdat);
extern void check_th_parms(FILE *fdat);
extern int bc_type(void);

#endif
