
/*******************************************************************************
*
* File admap.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Reordering of the gauge field variables in a form suitable for the FFT.
*
*   void gather_afld(sl3_alg_dble *ad,sl3_alg_dble **sad)
*     Gathers the field variables from the gauge potential ad and assigns
*     them to the array elements sad[mu][ix], mu=0,..,3, ix=0,..,VOLUME,
*     as described in the notes.
*
*   void scatter_afld(sl3_alg_dble **sad,sl3_alg_dble *ad)
*     Inverse of the operation performed by gather_afld().
*
* Gauge potentials are locally stored as described in main/README.global.
* The program gather_afld() assigns the field variable A_{mu}(x) to the
* array element sad[mu][ix], where
*
*  ix=x3+L3*x2+L2*L3*x1+L1*L2*L3*x0, 0<=x_mu<L_mu.
*
* While sad must be an array of size greater or equal to 4*VOLUME, the gauge
* potential ad is assumed to have size equal to at least 4*VOLUME+7*(BNDRY/4).
*
* The programs in this module act globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define ADMAP_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "sl3.h"
#include "aflds.h"
#include "global.h"

static int ofs[4],snu[4],init=0;


static void set_ofs(void)
{
   ofs[0]=VOLUME+(FACE0/2);
   ofs[1]=ofs[0]+(FACE0/2)+(FACE1/2);
   ofs[2]=ofs[1]+(FACE1/2)+(FACE2/2);
   ofs[3]=ofs[2]+(FACE2/2)+(FACE3/2);

   snu[0]=0;
   snu[1]=snu[0]+(FACE0/2);
   snu[2]=snu[1]+(FACE1/2);
   snu[3]=snu[2]+(FACE2/2);

   init=1;
}


static int offset(int ix,int mu)
{
   int iy,ib;

   if (ix<(VOLUME/2))
   {
      iy=iup[ix][mu];

      if (iy<VOLUME)
         return 8*(iy-(VOLUME/2))+2*mu+1;
      else
      {
         ib=iy-ofs[mu]-(BNDRY/2);

         return 4*VOLUME+snu[mu]+ib;
      }
   }
   else
      return 8*(ix-(VOLUME/2))+2*mu;
}


void gather_afld(sl3_alg_dble *ad,sl3_alg_dble **sad)
{
   int ix,iy,mu;

   if (init==0)
      set_ofs();

   gather_bnd_ad0(ad);

   for (ix=0;ix<VOLUME;ix++)
   {
      iy=ipt[ix];

      for (mu=0;mu<4;mu++)
         sad[mu][ix]=ad[offset(iy,mu)];
   }
}


void scatter_afld(sl3_alg_dble **sad,sl3_alg_dble *ad)
{
   int ix,iy,mu;

   if (init==0)
      set_ofs();

   for (ix=0;ix<VOLUME;ix++)
   {
      iy=ipt[ix];

      for (mu=0;mu<4;mu++)
         ad[offset(iy,mu)]=sad[mu][ix];
   }

   scatter_bnd_ad0(ad);
}
