
/*******************************************************************************
*
* File error.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Error handling functions.
*
*   void error(int test,int no,char *name,char *format,...)
*     Checks whether "test"=0 and if not aborts the program gracefully
*     with error number "no" after printing the "name" of the calling
*     program and an error message to stdout. The message is formed using
*     the "format" string and any additional arguments, exactly as in a
*     printf statement.
*
*   void error_root(int test,int no,char *name,char *format,...)
*     Same as error(), provided for compatibility.
*
*   void error_loc(int test,int no,char *name,char *format,...)
*     Same as error(), provided for compatibility.
*
*******************************************************************************/

#define ERROR_C

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "utils.h"


void error(int test,int no,char *name,char *format,...)
{
   va_list args;

   if (test!=0)
   {
      printf("\nError in %s (error no=%d):\n",name,no);
      va_start(args,format);
      vprintf(format,args);
      va_end(args);
      printf("\nProgram aborted\n");
      exit(no);
   }
}


void error_root(int test,int no,char *name,char *format,...)
{
   va_list args;

   if (test!=0)
   {
      printf("\nError in %s (error no=%d):\n",name,no);
      va_start(args,format);
      vprintf(format,args);
      va_end(args);
      printf("\nProgram aborted\n");
      exit(no);
   }
}


void error_loc(int test,int no,char *name,char *format,...)
{
   va_list args;

   if (test!=0)
   {
      printf("\nError in %s (error no=%d):\n",name,no);
      va_start(args,format);
      vprintf(format,args);
      va_end(args);
      printf("\nProgram aborted\n");
      exit(no);
   }
}
