
/*******************************************************************************
*
* File uflds.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Conversion of gauge potentials to SL(3,C) link fields.
*
*   void a2uflds(int n,sl3_alg_dble **ad,su3_dble *ud)
*     Expands exp{g0*ad[0][i]+g0^2*ad[1][i]+..+g0^(n+1)*ad[n][i]} in powers
*     of g0 and assigns the coefficient of g0^(k+1) to ud[l*(n+1)+k], where
*     k=0,..,n and l=0,..,4*VOLUME-1 labels the links on the lattice.
*
* The gauge potentials ad[0],..,ad[n] and the link field ud are assumed to
* have the size of the fields returned by reserve_wad(1) and reserve_wud(n+1),
* respectively (see utils/wspace.c).
*
* At global time 0 and T, a2uflds() sets the boundary values of the link
* field ud as required by the chosen boundary conditions.
*
* The program a2uflds() does not perform any communications and can be locally
* called.
*
*******************************************************************************/

#define UFLDS_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "pmdb.h"
#include "lattice.h"
#include "linalg.h"
#include "aflds.h"
#include "uflds.h"
#include "global.h"

static int ms=0;
static sl3_alg_dble **Aw,*Xw;


static void alloc_arrays(int n)
{
   int m,i;

   if (ms>0)
   {
      free(Aw);
      afree(Xw);
   }

   m=n+1;
   Aw=malloc(m*sizeof(*Aw));
   Xw=amalloc(m*sizeof(*Xw),ALIGN);

   error_loc((Aw==NULL)||(Xw==NULL),1,"alloc_arrays [uflds.c]",
             "Unable to allocate auxiliary arrays");

   for (i=0;i<m;i++)
      Aw[i]=NULL;

   set_calg2zero(m,Xw);
   ms=m;
}


void a2uflds(int n,sl3_alg_dble **ad,su3_dble *ud)
{
   int bc,m,ix,t,ifc,i;

   bc=bc_type();
   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   for (i=0;i<m;i++)
      Aw[i]=ad[i];

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            for (i=0;i<m;i++)
               Xw[i]=Aw[i][0];

            pexp_sl3_alg(m,Xw,ud);
         }
         else
            cm3x3_zero(m,ud);

         for (i=0;i<m;i++)
            Aw[i]+=1;

         ud+=m;
      }
   }

   if (cpr[0]==(NPROC0-1))
      cm3x3_zero(3*m,ud+7*(BNDRY/4)*m);
}
